@@ -9,13 +9,13 @@ from database import db
 
 def get_user_pass(login: str) -> Optional[str]:
     res = db.get_user_data(login=login)
-    return res[1] if res else None
+    return res[2] if res else None
 
 
 def user_in_group(login: str, group: str) -> bool:
     res = db.get_user_data(login=login)
     # bool cast needed for SQLite…
-    return bool(res[2]) if res and group == "ADMIN" else res is not None
+    return bool(res[3]) if res and group == "ADMIN" else res is not None
 
 
 def check_user_access(login: str, oid: str, mode) -> bool:
@@ -26,4 +26,4 @@ def check_user_access(login: str, oid: str, mode) -> bool:
 def init_app(app: Flask):
     app.get_user_pass(get_user_pass)
     app.user_in_group(user_in_group)
-    app.register_object_perms("users", check_user_access)
+    app.object_perms("users", check_user_access)
