@@ -3,12 +3,13 @@
 # FIXME tests are not perfectly isolated as they should be…
 #
 
+import re
 import pytest
 import App
 from App import app
 
 import FlaskSimpleAuth as fsa
-from FlaskSimpleAuth import Response, ConfigError
+from FlaskSimpleAuth import Response, ConfigError, ErrorResponse
 import json
 
 import AppExt
@@ -92,6 +93,13 @@ def pop_auth(app):
     app._auth, app._token, app._carrier, app._name = d["a"], d["t"], d["c"], d["n"]
     d.clear()
 
+def auth_header_basic(user: str):
+    from requests.auth import _basic_auth_str as basic_auth
+    return {"Authorization": basic_auth(user, App.UP[user])}
+
+def auth_header_token(user: str):
+    return {"Authorization": "Bearer " + app.create_token(user)}
+
 # test all auth variants on GET
 def all_auth(client, user, pswd, check, *args, **kwargs):
     # fake login
@@ -223,15 +231,31 @@ def test_register(client):
 def test_fsa_token():
     tsave, hsave = app._fsa._token, app._fsa._algo
     app._fsa._token, app._fsa._algo = "fsa", "blake2s"
+    app._fsa._issuer = "self"
+    foo_token = app.create_token("foo")
+    assert foo_token.startswith("Test/self:foo:")
+    assert app._fsa._get_any_token_auth(foo_token) == "foo"
+    app._fsa._issuer = None
     calvin_token = app.create_token("calvin")
     assert calvin_token[:12] == "Test:calvin:"
     assert app._fsa._get_any_token_auth(calvin_token) == "calvin"
+    # malformed token
+    try:
+        user = app._fsa._get_any_token_auth("not an FSA token")
+        assert False, "expecting a malformed error"
+    except fsa.ErrorResponse as e:
+        assert "invalid fsa token" in str(e)
     # bad timestamp format
     try:
         user = app._fsa._get_any_token_auth("R:U:demain:signature")
         assert False, "expecting a bad timestamp format"
     except fsa.ErrorResponse as e:
         assert "unexpected timestamp format" in e.message
+    try:
+        user = app._fsa._get_any_token_auth("Test:calvin:20201500000000:signature")
+        assert False, "expecting a bad timestamp format"
+    except fsa.ErrorResponse as e:
+        assert "unexpected limit" in e.message
     # force expiration
     grace = app._fsa._grace
     app._fsa._grace = -1000000
@@ -282,13 +306,15 @@ def test_jwt_token():
     tsave, hsave, app._fsa._token, app._fsa._algo = app._fsa._token, app._fsa._algo, "jwt", "HS256"
     Ksave, ksave = app._fsa._secret, app._fsa._sign
     # hmac signature scheme
+    app._fsa._issuer = "self"
     moe_token = app.create_token("moe")
     assert "." in moe_token and len(moe_token.split(".")) == 3
     user = app._fsa._get_any_token_auth(moe_token)
     assert user == "moe"
-    # again for caching
+    # again for caching test
     user = app._fsa._get_any_token_auth(moe_token)
     assert user == "moe"
+    app._fsa._issuer = None
     # expired token
     delay, grace = app._fsa._delay, app._fsa._grace
     app._fsa._delay, app._fsa._grace = -1, 0
@@ -340,26 +366,133 @@ def test_invalid_token():
         assert e.status == 401
 
 def test_password_check(client):
-    app._fsa._init_password_manager()
+    fsa = app._fsa
+    # standard password
+    fsa._init_password_manager()
     ref = app.hash_password("hello")
     assert app.check_password("hello", ref)
     assert not app.check_password("bad-pass", ref)
-    push_auth(app._fsa, ["password"])
+    # password alternate hook
+    assert fsa._password_check is None
+    def test_check_pass(user, pwd):
+        if user == "calvin" and pwd == "hobbes":
+            return True
+        elif pwd == "magic":
+            return True
+        elif pwd == "none":
+            return False
+        elif pwd == "Error":
+             raise ErrorResponse("test_check_pass error", 400)
+        else:
+             raise Exception("oops!")
+    fsa._password_check = test_check_pass
+    assert fsa._check_password("calvin", "hobbes") == "calvin"
+    assert fsa._check_password("susie", "magic") == "susie"
+    assert fsa._check_password("moe", "magic") == "moe"
+    try:
+        fsa._check_password("boo", "none")
+        assert False, "should have raised an error"
+    except ErrorResponse as e:
+        assert True, "none password was rejected"
+    try:
+        fsa._check_password("dad", "Error")
+        assert False, "should raise an error"
+    except ErrorResponse as e:
+        assert "test_check_pass error" in str(e)
+    try:
+        fsa._check_password("baa", "whatever")
+        assert False, "should raise an Exception"
+    except ErrorResponse as e:
+        assert "no such user" in str(e)
+    saved = fsa._get_user_pass
+    fsa.get_user_pass(None)
+    try:
+        fsa._check_password("calvin", "Oops!")
+        assert False, "should raise an error"
+    except ErrorResponse as e:
+        assert "invalid user/password" in str(e)
+    fsa.get_user_pass(saved)
+    fsa.password_check(None)
+    # password, through requests
+    push_auth(fsa, ["password"])
     res = check(401, client.get("/read", data={"USER": "dad", "PASS": "bad-dad-password"}))
     assert b"invalid password for" in res.data
     res = check(401, client.get("/read", data={"USER": "dad"}))
     assert b"missing password parameter" in res.data
-    pop_auth(app._fsa)
-    push_auth(app._fsa, ["basic"])
+    pop_auth(fsa)
+    # basic, through requests
+    push_auth(fsa, ["basic"])
     res = check(401, client.get("/read", headers={"Authorization": "Basic !!!"}))
     assert b"decoding error on authorization" in res.data
-    pop_auth(app._fsa)
-    pm = app._fsa._pm
-    app._fsa._pm = None
+    pop_auth(fsa)
+    # try plaintext
+    pm = fsa._pm
+    fsa._pm = None
     app.config.update(FSA_PASSWORD_SCHEME = "plaintext")
-    app._fsa._init_password_manager()
+    fsa._init_password_manager()
     assert app.hash_password("hello") == "hello"
-    app._fsa._pm = pm
+    fsa._pm = pm
+
+def test_password_quality():
+    fsa = app._fsa
+    # password len
+    assert fsa._password_len == 0
+    assert app.hash_password("") is not None
+    assert app.hash_password("c") is not None
+    assert app.hash_password("cy") is not None
+    fsa._password_len = 1
+    try:
+        app.hash_password("")
+        assert False, "len must be rejected"
+    except ErrorResponse as e:
+        assert "too short" in str(e)
+    assert app.hash_password("c") is not None
+    assert app.hash_password("cy") is not None
+    fsa._password_len = 2
+    assert app.hash_password("cy") is not None
+    # password re, eg password must contain a lc and uc letter
+    assert len(fsa._password_re) == 0
+    fsa._password_re = [
+        re.compile(r"[a-z]").search,
+        re.compile(r"[A-Z]").search,
+    ]
+    assert app.hash_password("Cy") is not None
+    try:
+        app.hash_password("CY")
+        assert False, "must detect missing lc letter"
+    except ErrorResponse as e:
+        assert "a-z" in str(e)
+    try:
+        app.hash_password("cy")
+        assert False, "must detect missing uc letter"
+    except ErrorResponse as e:
+        assert "A-Z" in str(e)
+    # password quality return
+    assert fsa._password_quality is None
+    def password_quality_checker(pwd):
+        if pwd == "G0od!":
+            return True
+        elif pwd == "B@d":
+            return False
+        else:
+            raise Exception(f"password quality checker is not happy about {pwd}")
+    fsa._password_quality = password_quality_checker
+    assert app.hash_password("G0od!") is not None
+    try:
+        app.hash_password("B@d")
+        assert False, "password should be rejected"
+    except ErrorResponse as e:
+        assert True, "password was rejected as expected"
+    # password quality exception
+    try:
+        app.hash_password("@ny-Password!")
+        assert False, "password should be rejected"
+    except ErrorResponse as e:
+        assert "not happy" in str(e)
+    # reset password checking rules
+    fsa._password_len = 0
+    fsa._password_re = []
+    fsa.password_quality(None)
 
 def test_authorize():
     assert app._fsa._user_in_group("dad", App.ADMIN)
@@ -428,6 +561,8 @@ def test_typed_params(client):
     assert int(res.data) == 42
     res = check(200, client.get("/request"))
     assert res.data.endswith(b"/request")
+    res = check(200, client.get("/special"))
+    assert res.data == b"foo-bla"
 
 def test_types(client):
     res = check(200, client.get("/type", data={"f": "1.0"}))
@@ -639,7 +774,7 @@ def test_something_3(client3):
 
 def test_401_redirect():
     import AppFact as af
-    app = af.create_app(FSA_401_REDIRECT="/login-page")
+    app = af.create_app(FSA_401_REDIRECT="/login-page", FSA_LOCAL="process")
     with app.test_client() as client:
         res = check(307, client.get("/something"))
         assert "/login-page" in res.location
@@ -658,142 +793,6 @@ def test_string(client):
     res = check(200, client.get("/string/foo"))
     assert res.data == b"foo"
 
-def test_reference():
-    v1, v2 = "hello!", "world!"
-    r1 = fsa.Reference()
-    r1.set(v1)
-    assert r1 == v1 and not r1 != v1
-    assert r1.startswith("hell")
-    r2 = fsa.Reference(set_name="set_object")
-    r2.set_object(v2)
-    assert r2 == v2
-    assert r2.endswith("ld!")
-    r3 = fsa.Reference("1")
-    assert r3 == "1" and r3 != "one"
-    assert r3.isdigit()
-    assert repr("1") == repr(r3)
-    # thread local stuff
-    def gen_data(i):
-        return f"data: {i}"
-    r = fsa.Reference(fun=gen_data, max_size=None)
-    assert r._nobjs == 0
-    assert isinstance(r.__hash__(), int)
-    assert r._nobjs == 1
-    assert isinstance(r._local.obj, str)
-    assert r == "data: 0"
-    # another thread
-    import threading
-    def run(i):
-        assert r == f"data: {i}"
-    t1 = threading.Thread(target=run, args=(1,))
-    t1.start()
-    t1.join()
-    assert r._nobjs == 2
-    t2 = threading.Thread(target=run, args=(2,))
-    t2.start()
-    t2.join()
-    assert r._nobjs == 3
-    # local one is still ok
-    assert r == "data: 0"
-    # FIXME thread objects are not really returned?
-    # error
-    try:
-        r = fsa.Reference(obj="hello", fun=gen_data)
-        assert False, "should have raised an exception"
-    except Exception as e:
-        assert "Reference cannot set both obj and fun" in str(e)
-    try:
-        r = fsa.Reference(set_name="add")
-        r.add()
-        assert False, "missing parameter in previous call"
-    except Exception as e:
-        assert "Reference must set either obj or fun" in str(e)
-    # auto thread
-    r = fsa.Reference()
-    r._set_fun(lambda i: i)
-    assert str(r) == "0"
-    # versatile
-    r = fsa.Reference(mode=fsa.Reference.Mode.VERSATILE)
-    r._set_fun(lambda i: i+10)
-    assert str(r) == "10"
-
-
-def test_ref_pool():
-    ref = fsa.Reference(fun=lambda i: i, max_size=2)
-    i = ref._get_obj()
-    assert len(ref._pool._avail) == 0
-    assert len(ref._pool._using) == 1
-    assert ref._pool._nobjs == 1
-    ref._ret_obj()
-    i = ref._get_obj()
-    assert len(ref._pool._avail) == 0
-    assert len(ref._pool._using) == 1
-    assert ref._pool._nobjs == 1
-    # this should return the same object as we are in the same thread
-    j = ref._get_obj()
-    assert i == j
-    ref._ret_obj()
-    assert len(ref._pool._avail) == 1
-    assert len(ref._pool._using) == 0
-    assert ref._pool._nobjs == 1
-    # test with 2 ordered threads to grasp to objects from the pool
-    import threading
-    event = threading.Event()
-    def run_1(i: int):
-        r = str(ref)      # get previous object #0
-        event.set()
-        assert r == str(i)
-        # ref._ret_obj()  # NOT RETURNED TO POOL
-    def run_2(i: int):
-        event.wait()
-        r = str(ref)      # generate a new object #1
-        assert r == str(i)
-        # ref._ret_obj()  # NOT RETURNED TO POOL
-    t1 = threading.Thread(target=run_1, args=(0,))
-    t2 = threading.Thread(target=run_2, args=(1,))
-    t1.start()
-    t2.start()
-    t1.join()
-    t2.join()
-    # use a 3rd reference to raise an pool max size exception
-    try:
-        ref._get_obj()   # failed attempt at generating #2
-        assert False, "must reach max_size"
-    except Exception as e:
-        assert "pool max size reached" in str(e)
-
-
-def test_pool():
-    # test max_use
-    p1 = fsa._Pool(fun = lambda i: i, max_size = 1, max_use = 2)
-    i = p1.get()
-    assert i == 0
-    p1.ret(i)
-    i = p1.get()
-    assert i == 0
-    p1.ret(i)
-    i = p1.get()
-    assert i == 1
-    p1.ret(i)
-    p1.__delete__()
-    # test with close and None
-    class T:
-        def __init__(self, count):
-            self._count = count
-        def close(self):
-            self._count = None
-            raise Exception("Oops!")
-        def __str__(self):
-            return f"T({self._count})"
-    p2 = fsa._Pool(fun = T, max_size = None, max_use = 1)
-    t = p2.get()
-    assert str(t) == "T(0)"
-    p2.ret(t)
-    t = p2.get()
-    assert str(t) == "T(1)"
-    p2.ret(t)
-    p2.__delete__()
-
 
 def test_www_authenticate(client):
     push_auth(app._fsa, "param")
@@ -921,7 +920,7 @@ def test_per_route(client):
 def test_bad_app():
     from AppBad import create_app
     # working versions, we basically test that there is no exception
-    app = create_app(FSA_AUTH="basic")
+    app = create_app(FSA_AUTH="basic", FSA_LOCAL="werkzeug")
     app = create_app(FSA_AUTH=["token", "basic"])
     app = create_app(auth="fake")
     app = create_app(auth=["token", "fake"])
@@ -990,6 +989,12 @@ def test_bad_app():
         assert False, "bad app creation must fail"
     except ConfigError as e:
         assert "bad" in str(e), "ok, bad app creation has failed"
+    # bad local
+    try:
+        app = create_app(FSA_LOCAL="oops!")
+        assert False, "bad app creation must fail"
+    except ConfigError as e:
+        assert "FSA_LOCAL" in str(e)
 
 class PK():
     def __init__(self, kind):
@@ -1236,6 +1241,27 @@ def test_group_errors():
     res = c.get("/float", data={"LOGIN": "calvin"})
     assert res.status_code == 500 and b"internal error with user_in_group" in res.data
 
+def test_scope_errors():
+    import AppFact as af
+    try:
+        app = af.create_app(FSA_AUTH="oauth", FSA_TOKEN_TYPE="fsa")
+        assert False, "should raise an exception"
+    except ConfigError as e:
+        assert "oauth" in str(e) and "JWT" in str(e)
+    try:
+        app = af.create_app(FSA_AUTH="oauth", FSA_TOKEN_TYPE="jwt", FSA_TOKEN_ISSUER=None)
+        assert False, "should raise an exception"
+    except ConfigError as e:
+        assert "oauth" in str(e) and "ISSUER" in str(e)
+    try:
+        app = af.create_app(FSA_AUTH="token", FSA_TOKEN_TYPE="fsa", FSA_TOKEN_ISSUER="god")
+        @app.get("/foo/bla", authorize=["read"], auth=["oauth"])
+        def get_foo_bla():
+            return "", 200
+        assert False, "should be rejected"
+    except ConfigError as e:
+        assert "JWT" in str(e)
+
 # run some checks on AppFact, repeat to exercise caching
 def run_some_checks(c, n=10):
     assert n >= 1
@@ -1333,6 +1359,7 @@ def test_warnings_and_errors():
     # unused length warning
     app = af.create_app(
         FSA_TOKEN_TYPE="jwt",
+        FSA_TOKEN_ISSUER="calvin",
         FSA_TOKEN_LENGTH=8,  # no used if jwt
         FSA_GET_USER_PASS=bad_gup_2,
     )
@@ -1351,6 +1378,8 @@ def test_warnings_and_errors():
     assert app._fsa._casts["foo"] != cast_foo
     s = "Hello World!"
     assert app._fsa._casts["foo"](s) == cast_foo(s)
+    # OAuth2 warnings
+    app = af.create_app(FSA_TOKEN_SCOPE=True)
     # type errors
     try:
         app = af.create_app(FSA_CAST="not a dict")
@@ -1362,6 +1391,11 @@ def test_warnings_and_errors():
         assert False, "should not get through"
     except ConfigError as e:
         assert "FSA_OBJECT_PERMS must be a dict" in str(e)
+    try:
+        app = af.create_app(FSA_SPECIAL_PARAMETER="not a dict")
+        assert False, "should not get through"
+    except ConfigError as e:
+        assert "FSA_SPECIAL_PARAMETER must be a dict" in str(e)
 
 def test_jsondata(client):
     # simple types, anything but strings
@@ -1409,3 +1443,81 @@ def test_jsondata(client):
     assert res.status_code == 200 and res.data == b'list: [false, true, [3, 14.0], {"q": 4}]'
     res = client.get("/json", json={"j": {"a": {"b": {"c": 3}}}})
     assert res.status_code == 200 and res.data == b'dict: {"a": {"b": {"c": 3}}}'
+
+def test_www_authenticate_priority(client):
+    # save current status
+    fsa = app._fsa
+    token, carrier, name = fsa._token, fsa._carrier, fsa._name
+    fsa._token, fsa._carrier, fsa._name = "fsa", "bearer", "Bearer"
+    BASIC = auth_header_basic("calvin")
+    TOKEN = auth_header_token("calvin")
+    # /perm/basic only basic
+    res = check(401, client.get("/perm/basic"))
+    assert "WWW-Authenticate" in res.headers
+    assert "Basic" in res.headers["WWW-Authenticate"]
+    res = check(200, client.get("/perm/basic", headers=BASIC))
+    res = check(401, client.get("/perm/basic", headers=TOKEN))
+    # /perm/token only token
+    res = check(401, client.get("/perm/token"))
+    assert "WWW-Authenticate" in res.headers
+    assert "Bearer" in res.headers["WWW-Authenticate"]
+    res = check(401, client.get("/perm/token", headers=BASIC))
+    res = check(200, client.get("/perm/token", headers=TOKEN))
+    # /perm/basic-token both ok
+    res = check(401, client.get("/perm/basic-token"))
+    assert "WWW-Authenticate" in res.headers
+    assert "Basic" in res.headers["WWW-Authenticate"]
+    res = check(200, client.get("/perm/basic-token", headers=BASIC))
+    res = check(200, client.get("/perm/basic-token", headers=TOKEN))
+    # /perm/token-basic both ok
+    res = check(401, client.get("/perm/token-basic"))
+    assert "WWW-Authenticate" in res.headers
+    assert "Bearer" in res.headers["WWW-Authenticate"]
+    res = check(200, client.get("/perm/token-basic", headers=BASIC))
+    res = check(200, client.get("/perm/token-basic", headers=TOKEN))
+    # test with other name
+    fsa._name = "Foo"
+    res = check(401, client.get("/perm/token"))
+    assert "WWW-Authenticate" in res.headers
+    assert "Foo" in res.headers["WWW-Authenticate"]
+    res = check(401, client.get("/perm/token-basic"))
+    assert "WWW-Authenticate" in res.headers
+    assert "Foo" in res.headers["WWW-Authenticate"]
+    # restore
+    fsa._token, fsa._carrier, fsa._name = token, carrier, name
+
+
+def test_jwt_authorization():
+    import AppFact as af
+    app = af.create_app(FSA_TOKEN_TYPE="jwt", FSA_REALM="comics", FSA_TOKEN_ISSUER="god")
+    # oauth in list auth
+    @app.get("/some/stuff", authorize=["read"], auth=["oauth"])
+    def get_some_stuff():
+        return "", 200
+    # config errors
+    try:
+        @app.get("/some/path", authorize=["read", "write"], auth=["oauth", "basic"])
+        def get_some_path():
+            return "", 200
+        assert False, "route should be rejected"
+    except fsa.ConfigError as e:
+        assert "mixed" in str(e)
+    app._fsa._issuer = None
+    try:
+        @app.patch("/any/stuff", authorize=["write"], auth="oauth")
+        def patch_any_stuff():
+            return "", 204
+        assert False, "route should be rejected"
+    except fsa.ConfigError as e:
+        assert "ISSUER" in str(e)
+    app._fsa._issuer = "god"
+    # usage errors
+    a = app._fsa
+    rosalyn_token = a._get_jwt_token(a._realm, "god", "rosalyn", 10.0, a._secret, scope=["character"])
+    rosalyn_auth=("Authorization", f"Bearer {rosalyn_token}")
+    moe_token = a._get_jwt_token(a._realm, "god", "moe", 10.0, a._secret, scope=["sidekick"])
+    moe_auth=("Authorization", f"Bearer {moe_token}")
+    client = app.test_client()
+    check(401, client.get("/perm/jwt-authz"))
+    check(200, client.get("/perm/jwt-authz", headers=[rosalyn_auth]))
+    check(403, client.get("/perm/jwt-authz", headers=[moe_auth]))
